-- Math3D.lua
Math3D = {}

-- Vector3 Functions
function Math3D.newVector(x, y, z)
    return {x = x or 0, y = y or 0, z = z or 0}
end

function Math3D.addVectors(v1, v2)
    return Math3D.newVector(v1.x + v2.x, v1.y + v2.y, v1.z + v2.z)
end

function Math3D.subtractVectors(v1, v2)
    return Math3D.newVector(v1.x - v2.x, v1.y - v2.y, v1.z - v2.z)
end

function Math3D.dotProduct(v1, v2)
    return v1.x * v2.x + v1.y * v2.y + v1.z * v2.z
end

function Math3D.crossProduct(v1, v2)
    return Math3D.newVector(
        v1.y * v2.z - v1.z * v2.y,
        v1.z * v2.x - v1.x * v2.z,
        v1.x * v2.y - v1.y * v2.x
    )
end

function Math3D.magnitude(v)
    return math.sqrt(v.x^2 + v.y^2 + v.z^2)
end

function Math3D.normalize(v)
    local mag = Math3D.magnitude(v)
    if mag == 0 then return Math3D.newVector(0, 0, 0) end
    return Math3D.newVector(v.x / mag, v.y / mag, v.z / mag)
end

function Math3D.multiplyVector(v, scalar)
    return Math3D.newVector(v.x * scalar, v.y * scalar, v.z * scalar)
end

function Math3D.lerpVectors(v1, v2, t)
    return Math3D.newVector(
        v1.x + (v2.x - v1.x) * t,
        v1.y + (v2.y - v1.y) * t,
        v1.z + (v2.z - v1.z) * t
    )
end

function Math3D.vectorToZXY(v)
    return Math3D.newVector(v.z, v.x, v.y)
end

function Math3D.vectorFromZXY(v)
    return Math3D.newVector(v.y, v.z, v.x)
end

function Math3D.vectorAngle(v1, v2)
    -- Compute the dot product of the two vectors
    local dotProduct = v1.x * v2.x + v1.y * v2.y + v1.z * v2.z
    
    -- Compute the magnitudes (lengths) of the vectors
    local magnitudeV1 = math.sqrt(v1.x^2 + v1.y^2 + v1.z^2)
    local magnitudeV2 = math.sqrt(v2.x^2 + v2.y^2 + v2.z^2)
    
    -- Compute the cosine of the angle
    local cosTheta = dotProduct / (magnitudeV1 * magnitudeV2)
    
    -- Clamp cosTheta to avoid floating point errors that could cause invalid acos input
    cosTheta = math.max(-1, math.min(1, cosTheta))
    
    -- Compute the angle in radians
    local angle = math.acos(cosTheta)
    
    return angle -- Return the angle in radians
end

-- Quaternion Functions

function Math3D.newQuaternion(w, x, y, z)
    return {w = w or 1, x = x or 0, y = y or 0, z = z or 0}
end

function Math3D.quaternionIdentity()
    return Math3D.newQuaternion(1, 0, 0, 0)
end

function Math3D.quaternionMultiply(q1, q2)
    return Math3D.newQuaternion(
        q1.w * q2.w - q1.x * q2.x - q1.y * q2.y - q1.z * q2.z,
        q1.w * q2.x + q1.x * q2.w + q1.y * q2.z - q1.z * q2.y,
        q1.w * q2.y - q1.x * q2.z + q1.y * q2.w + q1.z * q2.x,
        q1.w * q2.z + q1.x * q2.y - q1.y * q2.x + q1.z * q2.w
    )
end

function Math3D.quaternionConjugate(q)
    return Math3D.newQuaternion(q.w, -q.x, -q.y, -q.z)
end

function Math3D.quaternionMagnitude(q)
    return math.sqrt(q.w^2 + q.x^2 + q.y^2 + q.z^2)
end

function Math3D.normalizeQuaternion(q)
    local mag = Math3D.quaternionMagnitude(q)
    if mag == 0 then return Math3D.newQuaternion(1, 0, 0, 0) end
    return Math3D.newQuaternion(q.w / mag, q.x / mag, q.y / mag, q.z / mag)
end

function Math3D.quaternionAxisAngle(axis, angle)
    -- Normalize the rotation axis
    local normalizedAxis = Math3D.normalize(axis)
    
    -- Convert the angle to radians
    local halfAngle = angle * 0.5
    local sinHalfAngle = math.sin(halfAngle)
    
    -- Create the rotation quaternion
    local rotationQuaternion = Math3D.newQuaternion(
        math.cos(halfAngle),
        normalizedAxis.x * sinHalfAngle,
        normalizedAxis.y * sinHalfAngle,
        normalizedAxis.z * sinHalfAngle
    )
    
    -- Normalize the quaternion to avoid precision errors
    return Math3D.normalizeQuaternion(rotationQuaternion)
end


-- Rotate a vector using a quaternion
function Math3D.rotateVectorByQuaternion(v, q)
    local p = Math3D.newQuaternion(0, v.x, v.y, v.z)
    local qConj = Math3D.quaternionConjugate(q)
    local result = Math3D.quaternionMultiply(Math3D.quaternionMultiply(q, p), qConj)
    return Math3D.newVector(result.x, result.y, result.z)
end

function Math3D.quaternionToEuler(q)
    -- roll (x-axis rotation)
    local sinr_cosp = 2 * (q.w * q.x + q.y * q.z)
    local cosr_cosp = 1 - 2 * (q.x * q.x + q.y * q.y)
    local roll = math.atan2(sinr_cosp, cosr_cosp)

    -- pitch (y-axis rotation)
    local sinp = math.sqrt(1 + 2 * (q.w * q.y - q.x * q.z))
    local cosp = math.sqrt(1 - 2 * (q.w * q.y - q.x * q.z))
    local pitch = 2 * math.atan2(sinp, cosp) - math.pi / 2;

    -- yaw (z-axis rotation)
    local siny_cosp = 2 * (q.w * q.z + q.x * q.y)
    local cosy_cosp = 1 - 2 * (q.y * q.y + q.z * q.z)
    local yaw = math.atan2(siny_cosp, cosy_cosp)

    return pitch, yaw, roll
end

function Math3D.quaternionFromEuler(pitch, yaw, roll)
    -- Convert angles from degrees to radians if necessary
    pitch = pitch * 0.5
    yaw = yaw * 0.5
    roll = roll * 0.5

    -- Calculate sin and cos for half-angles
    local sinPitch = math.sin(pitch)
    local cosPitch = math.cos(pitch)
    local sinYaw = math.sin(yaw)
    local cosYaw = math.cos(yaw)
    local sinRoll = math.sin(roll)
    local cosRoll = math.cos(roll)

    -- Calculate quaternion components
    local w = cosRoll * cosPitch * cosYaw + sinRoll * sinPitch * sinYaw
    local x = sinRoll * cosPitch * cosYaw - cosRoll * sinPitch * sinYaw
    local y = cosRoll * sinPitch * cosYaw + sinRoll * cosPitch * sinYaw
    local z = cosRoll * cosPitch * sinYaw - sinRoll * sinPitch * cosYaw

    -- Return the quaternion
    return Math3D.newQuaternion(w, x, y, z)
end
