-- name: Galaxy planet prototype
-- description: A prototype of a round planet.\n\nMade by djoslin0, repurpose it for whatever...\nif you can stomache the math

E_MODEL_ICOSPHERE = smlua_model_util_get_id("icosphere_geo")
E_MODEL_SPACE = smlua_model_util_get_id("space_geo")
COL_ICOSPHERE = smlua_collision_util_get("icosphere_collision")

-- force sensible cameras and collisions
gLevelValues.fixCollisionBugs = 1
camera_set_romhack_override(RCO_ALL)

local icoscale = 0.75
local icoobj = nil
local icocenter = Math3D.newVector(0, 0, 0)
local icobaseradius = 1187
local icoradius = icobaseradius * icoscale
local coinobj = nil
local during_mario_update = false

local accumulatedRotation = Math3D.quaternionIdentity()
local prev_velocity = Math3D.newVector(0, 0, 0)
local did_gravity = false

local playerPosition = Math3D.newVector(0, 0, 0)
local playerRotation = Math3D.newVector(0, 0, 0)

local nonrotatingWorld = false

local frame = 0
local syncOffset = 100000

local gPlayerCache = {}
for i = 0, (MAX_PLAYERS - 1) do
    gPlayerCache[i] = {}
    gPlayerCache[i].position = Math3D.newVector(0, 0, 0)
    gPlayerCache[i].rotation = Math3D.newVector(0, 0, 0)
    gPlayerCache[i].lerp_position = Math3D.newVector(0, 0, 0)
    gPlayerCache[i].lerp_frame = 0
end

gGlobalSyncTable.scale = 0.75
gGlobalSyncTable.gravity = 1
gGlobalSyncTable.coinposx = 0
gGlobalSyncTable.coinposy = 0
gGlobalSyncTable.coinposz = 0

----------------------------------------

local function radians_to_sm64(radians)
    return math.floor(radians * 0x8000 / math.pi)
end

local function radians_from_sm64(sm64)
    return sm64 / 0x8000 * math.pi
end

local function radians_to_deg(radians)
    return math.floor(radians * 180 / math.pi)
end

function active_player(m)
    local np = gNetworkPlayers[m.playerIndex]
    if m.playerIndex == 0 then
        return true
    end
    if not np.connected then
        return false
    end
    return is_player_active(m)
end

---

local function apply_remote_transformation(m)
    -- this only runs on remote players
    -- in essence, our planet is rotated differently from other players
    -- so each player converts their position and rotation as if they were on a non-rotated planet
    -- and then we have to apply rotations based on our planet

    cache = gPlayerCache[m.playerIndex]

    -- since we had to disable the built in interpolation due to our worlds not matching,
    -- perform some basic position interpolation
    if cache.lerp_frame ~= frame then
        cache.lerp_position = Math3D.lerpVectors(cache.lerp_position, cache.position, 0.4)
        cache.lerp_frame = frame
    end

    -- rotate the remote player's position
    local npos = Math3D.newVector(cache.lerp_position.x, cache.lerp_position.y, cache.lerp_position.z)
    npos = Math3D.rotateVectorByQuaternion(Math3D.vectorToZXY(npos), accumulatedRotation)
    npos = Math3D.vectorFromZXY(npos)
    npos.x = npos.x + icocenter.x
    npos.y = npos.y + icocenter.y
    npos.z = npos.z + icocenter.z
    m.pos.x = npos.x
    m.pos.y = npos.y
    m.pos.z = npos.z
    m.peakHeight = m.pos.y
    obj_set_gfx_pos(m.marioObj, m.pos.x, m.pos.y, m.pos.z)

    -- rotate the remote player's rotation
    local conjugate = Math3D.quaternionConjugate(accumulatedRotation)
    local mQuaternion = Math3D.quaternionFromEuler(radians_from_sm64(cache.rotation.x), radians_from_sm64(cache.rotation.y), radians_from_sm64(cache.rotation.z))
    mQuaternion = Math3D.quaternionMultiply(accumulatedRotation, mQuaternion)
    mQuaternion = Math3D.normalizeQuaternion(mQuaternion)
    local pitch, yaw, roll = Math3D.quaternionToEuler(mQuaternion)
    m.faceAngle.x = radians_to_sm64(pitch)
    m.faceAngle.y = radians_to_sm64(yaw)
    m.faceAngle.z = radians_to_sm64(roll)
    obj_set_gfx_angle(m.marioObj,
        radians_to_sm64(pitch),
        radians_to_sm64(yaw),
        radians_to_sm64(roll))

    -- force disable our normal 'align-to-ground' matrix
    m.marioObj.header.gfx.throwMatrix = nil
end

local function force_gfx()
    -- this is called from the space skybox's update loop...
    -- the space skybox's loop runs after the planet and the players
    -- so we can update a bunch of graphics here after simulation on the players is done

    m = gMarioStates[0]

    -- the nonrotatingworld stuff is for debugging and can be removed
    if nonrotatingWorld then
        -- rotate player position
        local conjugate = Math3D.quaternionConjugate(accumulatedRotation)
        local difference = Math3D.newVector(playerPosition.x - icocenter.x, playerPosition.y - icocenter.y, playerPosition.z - icocenter.z)
        local rotated = Math3D.rotateVectorByQuaternion(Math3D.vectorToZXY(difference), conjugate)
        rotated = Math3D.vectorFromZXY(rotated)
        obj_set_gfx_pos(m.marioObj,
            icocenter.x + rotated.x,
            icocenter.y + rotated.y,
            icocenter.z + rotated.z)

        -- rotate player rotation
        local mQuaternion = Math3D.quaternionFromEuler(radians_from_sm64(m.faceAngle.x), radians_from_sm64(m.faceAngle.y), radians_from_sm64(m.faceAngle.z))
        mQuaternion = Math3D.quaternionMultiply(conjugate, mQuaternion)
        mQuaternion = Math3D.normalizeQuaternion(mQuaternion)
        local pitch, yaw, roll = Math3D.quaternionToEuler(mQuaternion)
        obj_set_gfx_angle(m.marioObj,
            radians_to_sm64(pitch),
            radians_to_sm64(yaw),
            radians_to_sm64(roll))
        m.marioObj.header.gfx.throwMatrix = nil
    end
    
    -- apply the rotations relative to our planet
    for i = 1, (MAX_PLAYERS - 1) do
        m = gMarioStates[i]
        if active_player(m) then
            apply_remote_transformation(m)
        end
    end
end

----------------------------------

local function bhv_icosphere_init(obj)
    obj.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
    obj.oCollisionDistance = 99999
    obj.collisionData = COL_ICOSPHERE
    cur_obj_scale(icoscale)
end

local function bhv_icosphere_loop(obj)
    -- apply the planet rotation
    local pitch, yaw, roll = Math3D.quaternionToEuler(accumulatedRotation)
    icoobj.oFaceAnglePitch = radians_to_sm64(pitch)
    icoobj.oFaceAngleYaw = radians_to_sm64(yaw)
    icoobj.oFaceAngleRoll = radians_to_sm64(roll)

    load_object_collision_model()

    -- this was for debugging, could be removed
    if nonrotatingWorld then
        icoobj.header.gfx.throwMatrix = nil
        icoobj.oFaceAnglePitch = 0
        icoobj.oFaceAngleYaw = 0
        icoobj.oFaceAngleRoll = 0
    end

    -- apply scale changes
    if gGlobalSyncTable.scale ~= icoscale then
        icoscale = gGlobalSyncTable.scale
        icoradius = icobaseradius * icoscale
        cur_obj_scale(icoscale)
    end
end

id_bhvIcosphere = hook_behavior(nil, OBJ_LIST_SURFACE, true, bhv_icosphere_init, bhv_icosphere_loop)

----------------------------------

local function bhv_space_init(obj)
    obj.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
end

local function bhv_space_loop(obj)
    -- the space skybox's loop runs after the planet and the players
    -- so we can update a bunch of graphics here after simulation on the players is done
    force_gfx()
end

id_bhvSpace = hook_behavior(nil, OBJ_LIST_DEFAULT, true, bhv_space_init, bhv_space_loop)

----------------------------------------

local function calculate_and_accumulate_rotation(m, icocenter, icoobj)
    -- this function performs the rotation on the planet in accordance to the player's movement direction

    -- sanity check
    if playerPosition.y < icocenter.y + icoradius * 0.9 or playerPosition.y > icocenter.y + icoradius * 4 then
        playerPosition = Math3D.newVector(icocenter.x, icocenter.y + icoradius, icocenter.z)
        m.pos.x = playerPosition.x
        m.pos.y = playerPosition.y
        m.pos.z = playerPosition.z
        m.peakHeight = m.pos.y
    end

    -- Compute the relative position of the player to the sphere center
    local up = Math3D.newVector(0, 1, 0)
    local difference = Math3D.newVector(m.pos.x - icocenter.x, m.pos.y - icocenter.y, m.pos.z - icocenter.z)
    local axis = Math3D.normalize(difference)
    local angle = Math3D.vectorAngle(difference, up)
    
    local crossed = Math3D.normalize(Math3D.crossProduct(axis, up))
    -- Get the rotation quaternion to align v with the up vector (0, 1, 0)
    local deltaQuaternion = Math3D.quaternionAxisAngle(Math3D.vectorToZXY(crossed), angle)

    -- Accumulate the rotation
    accumulatedRotation = Math3D.quaternionMultiply(deltaQuaternion, accumulatedRotation)
    accumulatedRotation = Math3D.normalizeQuaternion(accumulatedRotation)

    -- Convert to Euler angles
    local pitch, yaw, roll = Math3D.quaternionToEuler(accumulatedRotation)

    -- rotate the physical player position in accordance to how the planet rotated
    local ppos = Math3D.rotateVectorByQuaternion(difference, Math3D.quaternionAxisAngle(crossed, angle))
    playerPosition.x = icocenter.x + ppos.x
    playerPosition.y = icocenter.y + ppos.y
    playerPosition.z = icocenter.z + ppos.z

    -- save the physical player rotation so we can reapply it
    playerRotation.x = m.faceAngle.x
    playerRotation.y = m.faceAngle.y
    playerRotation.z = m.faceAngle.z

    -- perform gravity alterations
    if did_gravity and m.vel.y < prev_velocity.y then
        local distance = Math3D.magnitude(difference)
        m.vel.y = prev_velocity.y + (m.vel.y - prev_velocity.y) * (icoradius / distance) * gGlobalSyncTable.gravity
        did_gravity = false
    end
end

----------------------------------

local function before_mario_update_local(m)
    -- force our local position back to the physical world
    m.pos.x = playerPosition.x
    m.pos.y = playerPosition.y
    m.pos.z = playerPosition.z
    m.faceAngle.x = playerRotation.x
    m.faceAngle.y = playerRotation.y
    m.faceAngle.z = playerRotation.z
    m.peakHeight = m.pos.y
end

local function mario_update_local(m)
    if icoobj ~= nil then
        -- perform the magic
        calculate_and_accumulate_rotation(m, icocenter, icoobj)
    end
end

local function before_mario_update(m)
    during_mario_update = true
    if m.playerIndex == 0 then
        before_mario_update_local(m)
    else
        -- for remote players, check to see if we received an updated position
        -- updated positions are set to a very high y position (syncOffset)
        local cpos = gPlayerCache[m.playerIndex].position
        if active_player(m) and m.pos.y > (syncOffset - icoradius * 4) then
            -- store the received position/rotation and apply our rotations to it
            gPlayerCache[m.playerIndex].position = Math3D.newVector(m.pos.x, m.pos.y - syncOffset, m.pos.z)
            gPlayerCache[m.playerIndex].rotation = Math3D.newVector(m.faceAngle.x, m.faceAngle.y, m.faceAngle.z)
            apply_remote_transformation(m)
        end
    end
end

local function mario_update(m)
    if m.playerIndex == 0 then
        mario_update_local(m)
    end
    -- prevent players from constantly yelling from falling
    m.peakHeight = m.pos.y

    -- remember that we're currently not inside a mario update
    during_mario_update = false
end

local function mario_before_phys_step(m, stepType)
    if stepType == STEP_TYPE_AIR then
        if did_gravity == false and m.action ~= ACT_FLYING and m.action ~= ACT_BUBBLED then
            -- save our velocity before its altered in a phys step
            -- this way we can alter gravity later
            prev_velocity = Math3D.newVector(m.vel.x, m.vel.y, m.vel.z)
            did_gravity = true
        end
    end

    -- prevent remote players from playing the bonk sound all the time
    if m.playerIndex ~= 0 then
        if stepType == STEP_TYPE_AIR then
            return AIR_STEP_NONE
        end
        if stepType == STEP_TYPE_GROUND then
            return GROUND_STEP_NONE
        end
    end
end

local function update()
    frame = frame + 1

    m = gMarioStates[0]

    -- here we lie about the state of the local player
    -- we have to convert the position and rotation to something
    -- sensible that can be translated on the remote side

    -- rotate player position
    local conjugate = Math3D.quaternionConjugate(accumulatedRotation)
    local difference = Math3D.subtractVectors(playerPosition, icocenter)
    local rotated = Math3D.rotateVectorByQuaternion(Math3D.vectorToZXY(difference), conjugate)
    rotated = Math3D.vectorFromZXY(rotated)
    m.pos.x = rotated.x
    m.pos.y = rotated.y + syncOffset
    m.pos.z = rotated.z
    m.peakHeight = m.pos.y

    -- rotate player rotation
    local mQuaternion = Math3D.quaternionFromEuler(radians_from_sm64(playerRotation.x), radians_from_sm64(playerRotation.y), radians_from_sm64(playerRotation.z))
    mQuaternion = Math3D.quaternionMultiply(conjugate, mQuaternion)
    mQuaternion = Math3D.normalizeQuaternion(mQuaternion)
    local pitch, yaw, roll = Math3D.quaternionToEuler(mQuaternion)
    m.faceAngle.x = radians_to_sm64(pitch)
    m.faceAngle.y = radians_to_sm64(yaw)
    m.faceAngle.z = radians_to_sm64(roll)
end

local function on_level_init()
    icoobj = nil
    coinobj = nil
end

local function on_sync_valid()
    -- spawn the planet, the space skybox, and a debug coinobj
    if icoobj == nil then
        icocenter = Math3D.newVector(-2500, 15000, 2500)
        playerPosition = Math3D.newVector(icocenter.x, icocenter.y + icoradius * 1.1, icocenter.z)
        icoobj = spawn_non_sync_object(
            id_bhvIcosphere,
            E_MODEL_ICOSPHERE,
            icocenter.x, icocenter.y, icocenter.z,
            nil)
        spawn_non_sync_object(
            id_bhvSpace,
            E_MODEL_SPACE,
            icocenter.x, icocenter.y, icocenter.z,
            nil)
        coinobj = spawn_non_sync_object(
            id_bhvSpace,
            E_MODEL_KOOPA_SHELL,
            icocenter.x, icocenter.y, icocenter.z,
            nil)
        local m = gMarioStates[0]
        m.pos.x = playerPosition.x
        m.pos.y = playerPosition.y
        m.pos.z = playerPosition.z
        m.peakHeight = m.pos.y
    end
end

local function before_set_mario_action(m, action)
    if m.playerIndex ~= 0 and during_mario_update then
        -- prevent the remote players from changing their actions during mario_update
        -- we have to do this because they would constantly change into a falling or sliding
        -- state otherwise
        if action ~= m.action then
            return 1
        end
        m.wall = nil
    end
end

local function on_galaxy_size_command(msg)
    local num = tonumber(msg)
    if tonumber(msg) > 0 then
        if num >= 0.1 and num <= 3 then
            gGlobalSyncTable.scale = num
        end
        return true
    end
    return false
end

local function on_galaxy_gravity_command(msg)
    local num = tonumber(msg)
    if tonumber(msg) > 0 then
        if num >= 0.1 and num <= 3 then
            gGlobalSyncTable.gravity = tonumber(msg)
        end
        return true
    end
    return false
end

hook_event(HOOK_BEFORE_MARIO_UPDATE, before_mario_update)
hook_event(HOOK_MARIO_UPDATE, mario_update)
hook_event(HOOK_BEFORE_PHYS_STEP, mario_before_phys_step)
hook_event(HOOK_UPDATE, update)
hook_event(HOOK_ON_LEVEL_INIT, on_level_init)
hook_event(HOOK_ON_SYNC_VALID, on_sync_valid)
hook_event(HOOK_BEFORE_SET_MARIO_ACTION, before_set_mario_action)

if network_is_server() then
    hook_chat_command('galaxy-size', string.format("[0.1 - 3] sets planet size"), on_galaxy_size_command)
    hook_chat_command('galaxy-gravity', string.format("[0.1 - 3] sets planet gravity"), on_galaxy_gravity_command)
end